import flask
from datetime import datetime
from pydantic.dataclasses import dataclass
from typing import Any, List, Optional, Type, Sequence, Tuple
from ...local_log import LocalLogMessage, get_local_logs
from .workflows import make_stage_dto
from ...repositories.project.project import (
    ProjectRepository,
    WorkflowStage,
    FormStage,
)
from ...repositories.stage_run import (
    get_stage_run_repository,
    StageRunRepository,
    StageRun,
)


@dataclass
class StageCardContentItem:
    key: str
    value: Any
    type: str = "text"

    def to_dict(self) -> dict:
        return dict(
            key=self.key,
            value=self.value,
            type=self.type,
        )


StageCardContent = List[StageCardContentItem]


@dataclass
class StageRunCard:
    id: str
    created_at: datetime
    status: str
    content: StageCardContent

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            created_at=self.created_at.isoformat(),
            status=self.status,
            content=[item.to_dict() for item in self.content],
        )


@dataclass
class ColumnStage:
    id: str
    type: str
    path: Optional[str]
    title: str
    can_be_started: bool = False

    @staticmethod
    def create(stage: WorkflowStage) -> "ColumnStage":
        return ColumnStage(
            id=stage.id,
            type=stage.type_name,
            title=stage.title,
            path=stage.path if isinstance(stage, FormStage) else None,
            can_be_started=isinstance(stage, FormStage) and stage.is_initial,
        )

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            type=self.type,
            title=self.title,
            path=self.path,
            can_be_started=self.can_be_started,
        )


@dataclass
class KanbanColumn:
    selected_stage: ColumnStage
    stage_run_cards: List[StageRunCard]
    stages: List[ColumnStage]

    def to_dict(self) -> dict:
        return dict(
            selected_stage=self.selected_stage.to_dict(),
            stage_run_cards=[card.to_dict() for card in self.stage_run_cards],
            stages=[stage.to_dict() for stage in self.stages],
        )


@dataclass
class KanbanData:
    columns: List[KanbanColumn]
    next_stage_options: List[ColumnStage]

    def to_dict(self) -> dict:
        return dict(
            columns=[column.to_dict() for column in self.columns],
            next_stage_options=[option.to_dict() for option in self.next_stage_options],
        )


class KanbanController:
    def __init__(
        self,
        stage_run_repository: Type[StageRunRepository],
        project_repository: Type[ProjectRepository],
    ) -> None:
        self.stage_run_repository = stage_run_repository
        self.project_repository = project_repository

    def stage_run_content(self, stage_run: StageRun):
        visualization = self.project_repository.load().visualization
        visualization_defined_data = [
            StageCardContentItem(
                key=v.name,
                value=stage_run.data.get(v.name),
                type=v.type,
            )
            for v in visualization.items
        ]
        visualization_defined_keys = [v.name for v in visualization.items]
        extra_data = [
            StageCardContentItem(key=key, value=stage_run.data.get(key))
            for key in sorted(stage_run.data.keys())
            if key not in visualization_defined_keys
        ]
        return visualization_defined_data + extra_data

    def _get_next_stages(
        self, previous_stage: Optional[str] = None
    ) -> Sequence[WorkflowStage]:
        project = ProjectRepository.load()
        if previous_stage:
            origin = project.get_stage(previous_stage)
            if not origin:
                raise Exception(f"Stage {previous_stage} not found")

            next_stages = []
            for next_stage in origin.workflow_transitions:
                stage = project.get_stage(next_stage.target_id)
                if not stage:
                    raise Exception(f"Stage {next_stage.target_id} not found")
                next_stages.append(stage)

            return next_stages

        else:
            return project.get_initial_stages()

    def _get_column_stages(self, selected_stages_ids: List[str], column_idx: int):
        if column_idx == 0:
            return self._get_next_stages()
        else:
            return self._get_next_stages(selected_stages_ids[column_idx - 1])

    def get_data(self, selected_stages_ids: List[str]) -> KanbanData:
        project = self.project_repository.load()

        return KanbanData(
            columns=[
                KanbanColumn(
                    selected_stage=ColumnStage.create(
                        project.get_stage_raises(selected_stage)
                    ),
                    stages=[
                        ColumnStage.create(s)
                        for s in self._get_column_stages(
                            selected_stages_ids, column_idx
                        )
                    ],
                    stage_run_cards=[
                        StageRunCard(
                            id=stage_run.id,
                            status=stage_run.status,
                            created_at=stage_run.created_at,
                            content=self.stage_run_content(stage_run),
                        )
                        for stage_run in self.stage_run_repository.find_leaves(
                            {"stage": selected_stage}
                        )
                    ],
                )
                for column_idx, selected_stage in enumerate(selected_stages_ids)
            ],
            next_stage_options=[
                ColumnStage.create(s)
                for s in self._get_column_stages(
                    selected_stages_ids, len(selected_stages_ids)
                )
            ],
        )

    def get_ancestor_logs(
        self, stage_run_id: str
    ) -> List[Tuple[StageRun, List[LocalLogMessage]]]:
        ancestors = self.stage_run_repository.find_ancestors(stage_run_id)
        return [
            (ancestor, get_local_logs(ancestor.execution_id)) for ancestor in ancestors
        ]


def get_editor_bp():
    project_repository = ProjectRepository
    stage_run_repository = get_stage_run_repository()

    controller = KanbanController(stage_run_repository, project_repository)
    bp = flask.Blueprint("kanban", __name__)

    @bp.post("/")
    def _get_kanban():
        if flask.request.json is None:
            flask.abort(400)
        selected_stages_ids = flask.request.json.get("selected_stages_ids", [])
        return controller.get_data(selected_stages_ids).to_dict()

    @bp.get("/logs/<stage_run_id>")
    def _get_ancestor_logs(stage_run_id: str):
        def stage_from_run(stage_run: StageRun):
            stage = project_repository.load().get_action_raises(stage_run.stage)
            return make_stage_dto(stage)

        return [
            {
                "stage": stage_from_run(stage_run),
                "stage_run": stage_run.to_dto(),
                "logs": [log.json() for log in logs],
            }
            for stage_run, logs in controller.get_ancestor_logs(stage_run_id)
        ]

    return bp



import flask
from typing import Optional

from ...usage import usage
from ...repositories import StageRunRepository


class StageRunsController:
    def get_stage_runs(self, stage: Optional[str] = None):
        return [
            stage_run.to_dto()
            for stage_run in StageRunRepository.find_leaves(dict(stage=stage))
        ]


def get_editor_bp():
    controller = StageRunsController()
    bp = flask.Blueprint("editor_stage_runs", __name__)

    @bp.get("/")
    @usage
    def _get_stage_runs():
        stage = flask.request.args.get("stage")
        return controller.get_stage_runs(stage)

    return bp



import flask

from ...usage import usage

from ...utils.format import normalize_path


def get_editor_bp():
    bp = flask.Blueprint("editor_utils", __name__)

    @bp.post("/normalize_path")
    @usage
    def _normalize_path():
        data = flask.request.json
        if not data:
            flask.abort(400)

        path = data["path"]
        if not path:
            flask.abort(400)

        return normalize_path(path)

    return bp



from ..repositories.project.project import ActionStage
from .static_execution import StaticExecution
from .execution import RequestData


class ScriptExecution(StaticExecution):
    stage_run_id: str

    def __init__(self, stage: ActionStage, stage_run_id: str):
        self.stage_run_id = stage_run_id

        request = RequestData(
            headers={},
            body="",
            method="GET",
            query_params={},
        )

        super().__init__(stage=stage, is_initial=False, request=request)

    def setup_context(self, request: RequestData):
        self.init_stage_run(self.stage_run_id)



import copy, uuid, requests
from datetime import datetime
from pydantic.dataclasses import dataclass
from typing import Optional, List, Dict, Type, Mapping, Any

from ..utils import real_datetime_fromisoformat
from ..utils.environment import SIDECAR_URL, SIDECAR_HEADERS

end_status = ["failed", "finished", "abandoned"]

status_transitions = {
    "waiting": ["running"],
    "running": end_status,
}


@dataclass
class StageRun:
    id: str
    stage: str
    data: dict
    status: str
    created_at: datetime
    parent_id: Optional[str]
    execution_id: Optional[str] = None

    def __getitem__(self, key):
        return self.data.get(key)

    def __setitem__(self, key, value):
        self.data[key] = value

    def get(self, key, default=None):
        return self.data.get(key, default)

    @property
    def is_end(self):
        return self.status in end_status

    @property
    def is_abandoned(self):
        return self.status == "abandoned"

    def to_dto(self):
        return copy.deepcopy(self.__dict__)

    def clone(self):
        return StageRun(**self.to_dto())


class StageRunRepository:
    valid_filter_keys = ["data", "stage", "status", "parent_id"]

    @classmethod
    def validate_filter_keys(cls, filter: Dict) -> None:
        for key in filter.keys():
            if key not in cls.valid_filter_keys:
                raise Exception(f"Invalid filter key {key}")

    @classmethod
    def hydrate_next_dto(cls, parent: StageRun, dtos: List[Dict]) -> List[Dict]:
        next_dtos = []

        for dto in dtos:
            stage = dto.get("stage")
            data = {**parent.data, **dto.get("data", {})}
            next_dtos.append(dict(stage=stage, data=data))

        return next_dtos

    @classmethod
    def find(cls, filter: Dict) -> List[StageRun]:
        raise NotImplementedError()

    @classmethod
    def find_ancestors(cls, id: str) -> List[StageRun]:
        raise NotImplementedError()

    @classmethod
    def find_leaves(cls, filter: Dict) -> List[StageRun]:
        raise NotImplementedError()

    @classmethod
    def get(cls, id: str) -> StageRun:
        raise NotImplementedError()

    @classmethod
    def change_state(
        cls, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        raise NotImplementedError()

    @classmethod
    def create_initial(cls, stage: str) -> StageRun:
        raise NotImplementedError()

    @classmethod
    def create_next(cls, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        raise NotImplementedError()


class LocalStageRunRepository(StageRunRepository):
    _stage_runs: List[StageRun] = []

    @classmethod
    def clear(cls):
        cls._stage_runs = []
        cls._locks = []

    @classmethod
    def get(cls, id: str) -> StageRun:
        for stage_run in cls._stage_runs:
            if stage_run.id == id:
                return stage_run

        raise Exception(f"StageRun with id {id} not found")

    @classmethod
    def find(cls, filter: Dict) -> List[StageRun]:
        cls.validate_filter_keys(filter)

        stage = filter.get("stage")
        data = filter.get("data")
        status = filter.get("status")
        parent_id = filter.get("parent_id")

        return [
            stage_run
            for stage_run in cls._stage_runs
            if (
                (not stage or stage_run.stage == stage)
                and (not data or stage_run.data == data)
                and (not status or stage_run.status == status)
                and (not parent_id or stage_run.parent_id == parent_id)
            )
        ]

    @classmethod
    def find_ancestors(cls, id: str) -> List[StageRun]:
        ancestors = []
        stage_run = cls.get(id)
        ancestors.append(stage_run)
        while stage_run.parent_id is not None:
            stage_run = cls.get(stage_run.parent_id)
            ancestors.insert(0, stage_run)
        return ancestors

    @classmethod
    def find_leaves(cls, filter: Dict) -> List[StageRun]:
        cls.validate_filter_keys(filter)
        filter_matches = cls.find(filter)
        parent_ids = set(u.parent_id for u in cls._stage_runs)

        return [
            stage_run for stage_run in filter_matches if stage_run.id not in parent_ids
        ]

    @classmethod
    def create_initial(cls, stage: str, data: dict = {}) -> StageRun:
        stage_run = StageRun(
            id=str(uuid.uuid4()),
            status="waiting",
            stage=stage,
            data=data,
            parent_id=None,
            created_at=datetime.now(),
        )

        cls._stage_runs.append(stage_run)

        return stage_run

    @classmethod
    def create_next(cls, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        created = []
        for dto in cls.hydrate_next_dto(parent, dtos):
            stage_run = StageRun(
                id=str(uuid.uuid4()),
                parent_id=parent.id,
                status="waiting",
                **dto,
                created_at=datetime.now(),
            )
            created.append(stage_run)
            cls._stage_runs.append(stage_run)
        return created

    @classmethod
    def change_state(
        cls, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        stage_run = cls.get(id)

        if not stage_run.execution_id and execution_id is not None:
            stage_run.execution_id = execution_id

        if stage_run.status in end_status:
            return False

        if status in status_transitions.get(stage_run.status, []):
            stage_run.status = status
            return True

        return False


class ProductionStageRunRepository(StageRunRepository):
    @classmethod
    def _request(
        cls,
        method: str,
        path: str,
        body: Any = None,
        params: dict = {},
        raise_for_status: bool = True,
    ):
        headers: Mapping[str, str] = SIDECAR_HEADERS
        r = requests.request(
            method=method,
            url=f"{SIDECAR_URL}/stage-runs{path}",
            headers=headers,
            json=body,
            params=params,
        )

        if raise_for_status:
            r.raise_for_status()

        return r

    @staticmethod
    def create_from_dto(dto: dict) -> StageRun:
        return StageRun(
            id=dto["id"],
            stage=dto["stage"],
            data=dto["data"],
            status=dto["status"],
            parent_id=dto["parentId"],
            created_at=real_datetime_fromisoformat(dto["createdAt"]),
        )

    @classmethod
    def get(cls, id: str) -> StageRun:
        r = cls._request("GET", path=f"/{id}")
        return cls.create_from_dto(r.json())

    @classmethod
    def find(cls, filter: Dict) -> List[StageRun]:
        cls.validate_filter_keys(filter)
        r = cls._request("GET", path="/", params=filter)
        return [cls.create_from_dto(dto) for dto in r.json()]

    @classmethod
    def find_ancestors(cls, id: str) -> List[StageRun]:
        r = cls._request("GET", path=f"/{id}/ancestors")
        return [cls.create_from_dto(dto) for dto in r.json()]

    @classmethod
    def create_initial(cls, stage: str) -> StageRun:
        body = dict(data={}, stage=stage)
        r = cls._request("POST", path="/", body=body)
        return cls.create_from_dto(r.json())

    @classmethod
    def create_next(cls, parent: StageRun, dtos: List[Dict]) -> List[StageRun]:
        if len(dtos) == 0:
            return []

        parent_id = parent.id
        next_dtos = cls.hydrate_next_dto(parent, dtos)
        r = cls._request("PUT", path=f"/{parent_id}/children", body=next_dtos)
        return [cls.create_from_dto(dto) for dto in r.json()]

    @classmethod
    def change_state(
        cls, id: str, status: str, execution_id: Optional[str] = None
    ) -> bool:
        r = cls._request(
            "PATCH", path=f"/{id}", body={"status": status}, raise_for_status=False
        )

        if r.status_code == 409:
            return False

        r.raise_for_status()

        return True


def get_stage_run_repository() -> Type[StageRunRepository]:
    if SIDECAR_URL is None:
        return LocalStageRunRepository
    else:
        return ProductionStageRunRepository



from abc import abstractmethod, ABC
from typing import Any, List, Union, Callable, Optional


class Widget(ABC):
    @abstractmethod
    def render(self, ctx: dict):
        raise NotImplementedError("render not implemented")

    @abstractmethod
    def set_props(self, props):
        raise NotImplementedError("set_props not implemented")


class Output(Widget):
    type: str

    def has_errors(self):
        return False


class Input(Widget):
    type: str
    errors: List[str] = (lambda: [])()
    empty_value: Any = None
    required: Optional[Union[bool, str]] = None

    def __init__(self, key: str) -> None:
        super().__init__()
        self.key = key

    def set_value(self, value, set_errors=False):
        old_value = self.value if hasattr(self, "value") else None
        self.value = value

        if set_errors and old_value != value:
            self.set_errors()

    def validate(self) -> List[str]:
        errors = []
        for validator in self.validators:
            errors += validator()
        return errors

    def set_errors(self):
        self.errors = self.validate()

    def has_errors(self):
        return len(self.errors) > 0

    def is_value_unset(self) -> bool:
        return (
            self.value == None
            or self.value == ""
            or self.value == []
            or (isinstance(self.empty_value, bool) and self.value == False)
            or self.value == {}
        )

    def _validate_required(self) -> List[str]:
        if self.required and self.is_value_unset():
            if type(self.required) == str:
                return [self.required]

            return ["i18n_error_required_field"]

        return []

    @property
    def validators(self) -> List[Callable[[], List[str]]]:
        return [self._validate_required]

    def parse_value(self, value):
        _value = value
        return _value if _value is not None else self.empty_value

    @abstractmethod
    def serialize_value(self):
        raise NotImplementedError("serialize_value not implemented")


class OptionalListInput(Input):
    # Input widgets that can be either a single value or a list of values
    multiple: bool = False
    min: Optional[int] = None
    max: Optional[int] = None

    def is_value_unset(self) -> bool:
        if self.multiple:
            return self.value == None
        else:
            return super().is_value_unset()

    def _single_value_or_list(self, value: List) -> Union[List, Any]:
        if not isinstance(value, list) and value is not None:
            raise Exception("Non-list value received")

        if value is None:
            return self.empty_value

        if self.multiple:
            return value or []

        return value[0] if len(value) > 0 else None

    def _validate_list_min_max(self) -> List[str]:
        if self.min is not None and len(self.value) < self.min:
            return ["i18n_error_min_list"]

        if self.max is not None and len(self.value) > self.max:
            return ["i18n_error_max_list"]

        return []

    @property
    def validators(self) -> List[Callable[[], List[str]]]:
        return super().validators + [self._validate_list_min_max]

    def parse_value(self, value: List) -> Union[List, Any]:
        _value = self._single_value_or_list(value)

        return super().parse_value(_value)


class NumberishInput(Input):
    min: Optional[Union[int, float]] = None
    max: Optional[Union[int, float]] = None

    @property
    def validators(self) -> List[Callable[[], List[str]]]:
        return super().validators + [self._validate_number_min_max]

    def _validate_number_min_max(self) -> List[str]:
        if type(self.value) != int and type(self.value) != float:
            return []

        if self.min is not None and self.value < self.min:
            return ["i18n_error_min_number"]

        if self.max is not None and self.value > self.max:
            return ["i18n_error_max_number"]

        return []



import flask

from ...usage import usage
from .main import MainController
from ..workflow_engine import workflow_engine
from ...execution.execution import RequestData
from ...execution.job_execution import JobExecution
from ...repositories.project.project import ProjectRepository
from ..utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_jobs", __name__)

    @bp.get("/<path:id>")
    @usage
    def _get_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)
        return job.editor_dto

    @bp.get("/")
    @usage
    def _get_jobs():
        return [f.editor_dto for f in controller.get_jobs()]

    @bp.post("/")
    @usage
    def _create_job():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        job = controller.create_job(title, file)
        return job.editor_dto

    @bp.put("/<path:id>")
    @usage
    def _update_runtime(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        job = controller.update_stage(id, data)
        return job.editor_dto if job else None

    @bp.delete("/<path:id>")
    @usage
    def _delete_job(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_job(id, remove_file)
        return {"success": True}

    @bp.post("/<path:id>/test")
    @usage
    def _test_job(id: str):
        job = controller.get_job(id)
        if not job:
            flask.abort(404)

        execution = JobExecution(job)
        execution.run()
        workflow_engine.notify_ran(execution)

        return {
            "stdout": "".join(execution.stdout if execution else []),
            "stderr": "".join(execution.stderr if execution else []),
        }

    return bp



import flask

from ...usage import usage
from .main import MainController
from ..workflow_engine import workflow_engine
from ...execution.script_execution import ScriptExecution
from ..utils import is_it_true


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_scripts", __name__)

    @bp.get("/<path:id>")
    @usage
    def _get_script(id: str):
        script = controller.get_script(id)
        if not script:
            flask.abort(404)
        return script.editor_dto

    @bp.get("/")
    @usage
    def _get_scripts():
        return [f.editor_dto for f in controller.get_scripts()]

    @bp.post("/")
    @usage
    def _create_script():
        data = flask.request.json
        if not data:
            flask.abort(400)
        title = data.get("title")
        file = data.get("file")
        if not title or not file:
            flask.abort(400)
        script = controller.create_script(title, file)
        return script.editor_dto

    @bp.put("/<path:id>")
    @usage
    def _update_script(id: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        script = controller.update_stage(id, data)
        return script.editor_dto if script else None

    @bp.delete("/<path:id>")
    @usage
    def _delete_script(id: str):
        remove_file = flask.request.args.get(
            "remove_file", default=False, type=is_it_true
        )
        controller.delete_script(id, remove_file)
        return {"success": True}

    @bp.post("/<path:id>/test")
    @usage
    def _test_script(id: str):
        script = controller.get_script(id)
        stage_run_id = flask.request.args.get("stage_run_id")

        if not script or not stage_run_id:
            flask.abort(404)

        execution = ScriptExecution(script, stage_run_id)
        execution.run()
        workflow_engine.notify_ran(execution)

        return {
            "stdout": "".join(execution.stdout if execution else []),
            "stderr": "".join(execution.stderr if execution else []),
        }

    return bp



import uuid

from .base_migration import Migration


class Migration001(Migration):
    @staticmethod
    def target_version() -> str:
        return "1.0"

    @staticmethod
    def source_version() -> str:
        return "0.1"

    def _add_visualization_settings(self) -> None:
        if "visualization" not in self.data:
            self.data["visualization"] = []

    def _remove_dashes(self) -> None:
        if "dashes" in self.data:
            if len(self.data["dashes"]):
                self.warnings.append(
                    "Dashes are no longer supported, from v1.18 onwards. Your Dashes were removed from the abstra.json file. Retrieve them from the backup file."
                )
            del self.data["dashes"]

    def _missing_entities(self) -> None:
        if "hooks" not in self.data:
            self.data["hooks"] = []

        if "jobs" not in self.data:
            self.data["jobs"] = []

        if "forms" not in self.data:
            self.data["forms"] = []

        if "scripts" not in self.data:
            self.data["scripts"] = []

        if "workspace" not in self.data:
            self.data["workspace"] = {
                "name": "Untitled Workspace",
                "sidebar": [],
            }

        if "workflow_transitions" not in self.data:
            self.data["workflow_transitions"] = []

    def _add_workflow_positions(self) -> None:
        size = 110
        forms = self.data["forms"]
        for idx, form in enumerate(forms):
            if "workflow_position" not in form:
                form["workflow_position"] = [0, idx * size]

        for idx, job in enumerate(self.data["jobs"]):
            if "workflow_position" not in job:
                job["workflow_position"] = [size, idx * size]

        for idx, hook in enumerate(self.data["hooks"]):
            if "workflow_position" not in hook:
                hook["workflow_position"] = [2 * size, idx * size]

    def _remove_is_initial(self) -> None:
        for form in self.data["forms"]:
            if "is_initial" in form:
                del form["is_initial"]

        for job in self.data["jobs"]:
            if "is_initial" in job:
                del job["is_initial"]

        for hook in self.data["hooks"]:
            if "is_initial" in hook:
                del hook["is_initial"]

        for script in self.data["scripts"]:
            if "is_initial" in script:
                del script["is_initial"]

    def _add_workflow_transitions(self) -> None:
        for form in self.data["forms"]:
            if "transitions" not in form:
                form["transitions"] = []

        for job in self.data["jobs"]:
            if "transitions" not in job:
                job["transitions"] = []

        for hook in self.data["hooks"]:
            if "transitions" not in hook:
                hook["transitions"] = []

        for script in self.data["scripts"]:
            if "transitions" not in script:
                script["transitions"] = []

    def _add_ids(self) -> None:
        for form in self.data["forms"]:
            if "id" not in form:
                form["id"] = str(uuid.uuid4())

        for job in self.data["jobs"]:
            job["id"] = job["identifier"]
            del job["identifier"]

        for hook in self.data["hooks"]:
            if "id" not in hook:
                hook["id"] = str(uuid.uuid4())

        for script in self.data["scripts"]:
            if "id" not in script:
                script["id"] = str(uuid.uuid4())

    def _get_id_by_path(self, path: str) -> dict:
        for form in self.data["forms"]:
            if form["path"] == path:
                return form["id"]

        for hook in self.data["hooks"]:
            if hook["path"] == path:
                return hook["id"]

        for script in self.data["scripts"]:
            if script["path"] == path:
                return script["id"]

        raise Exception(f"Could not find entity with path {path}")

    def _transition_paths_to_ids(self) -> None:
        for form in self.data["forms"]:
            for transition in form["transitions"]:
                if "target_id" not in transition:
                    transition["target_id"] = self._get_id_by_path(
                        transition["target_path"]
                    )
                    del transition["target_path"]

        for job in self.data["jobs"]:
            for transition in job["transitions"]:
                if "target_id" not in transition:
                    transition["target_id"] = self._get_id_by_path(
                        transition["target_path"]
                    )
                    del transition["target_path"]

        for hook in self.data["hooks"]:
            for transition in hook["transitions"]:
                if "target_id" not in transition:
                    transition["target_id"] = self._get_id_by_path(
                        transition["target_path"]
                    )
                    del transition["target_path"]

        for script in self.data["scripts"]:
            for transition in script["transitions"]:
                if "target_id" not in transition:
                    transition["target_id"] = self._get_id_by_path(
                        transition["target_path"]
                    )
                    del transition["target_path"]

    def _migrate(self) -> None:
        self._missing_entities()
        self._add_workflow_positions()
        self._remove_is_initial()
        self._add_workflow_transitions()
        self._add_visualization_settings()
        self._remove_dashes()
        self._add_ids()
        self._transition_paths_to_ids()
        self._bump_version()



from typing import Type, List, Dict

from ..repositories.stage_run import get_stage_run_repository, StageRunRepository
from ..execution.execution import Execution, UnsetStageRun
from ..execution.script_execution import ScriptExecution
from ..execution.job_execution import JobExecution
from ..repositories.stage_run import StageRun
from ..threaded import threaded

from ..repositories.project.project import (
    ActionStage,
    ProjectRepository,
    ConditionStage,
    IteratorStage,
    ControlStage,
    ScriptStage,
    JobStage,
)


class WorkflowEngine:
    def __init__(self, stage_run_repo: Type[StageRunRepository]):
        self.stage_run_repo = stage_run_repo

    def get_stage(self, id: str):
        # TODO: inject this
        project = ProjectRepository.load()
        return project.get_stage(id)

    def notify_ran(self, execution: Execution):
        if not execution.stage_run:
            raise UnsetStageRun()

        stage = execution.stage
        status = execution.status
        stage_run = execution.stage_run

        transition_type = f"{stage.type_name}s:{status}"
        next_stage_run_dtos = action_strategy(stage, stage_run, transition_type)
        self.pub(stage_run, next_stage_run_dtos)

    def pub(self, parent_stage_run: StageRun, stage_run_dtos: List[Dict]):
        for stage_run in self.stage_run_repo.create_next(
            parent_stage_run, stage_run_dtos
        ):
            self.consume(stage_run)

    @threaded
    def run_job(self, stage: JobStage):
        execution = JobExecution(stage)
        execution.run()
        self.notify_ran(execution)

    @threaded
    def run_script(self, stage: ScriptStage, stage_run: StageRun):
        execution = ScriptExecution(stage, stage_run.id)
        execution.run()
        self.notify_ran(execution)

    @threaded
    def run_control(self, stage: ControlStage, stage_run: StageRun):
        next_stage_run_dtos: List[Dict] = []
        self.stage_run_repo.change_state(stage_run.id, "running")

        if isinstance(stage, ConditionStage):
            next_stage_run_dtos = condition_strategy(stage, stage_run)

        if isinstance(stage, IteratorStage):
            next_stage_run_dtos = iterator_strategy(stage, stage_run)

        self.stage_run_repo.change_state(stage_run.id, "finished")
        self.pub(stage_run, next_stage_run_dtos)

    def consume(self, stage_run: StageRun):
        stage = self.get_stage(stage_run.stage)
        if not stage:
            raise Exception(f"Stage {stage_run.stage} not found")

        if isinstance(stage, ScriptStage):
            return self.run_script(stage, stage_run)

        if isinstance(stage, (ConditionStage, IteratorStage)):
            return self.run_control(stage, stage_run)


def condition_strategy(
    stage: ConditionStage,
    stage_run: StageRun,
    transition_type="conditions:patterMatched",
):
    variable_value = stage_run.data.get(stage.variable_name)

    next_stage_ids = [
        t.target_id
        for t in stage.workflow_transitions
        if t.type == transition_type and t.condition_value == variable_value
    ]

    return [dict(stage=stage_id, data=stage_run.data) for stage_id in next_stage_ids]


def iterator_strategy(
    stage: IteratorStage, stage_run: StageRun, transition_type="iterators:each"
):
    variable_value = stage_run.data.get(stage.variable_name)

    if not isinstance(variable_value, list):
        return []

    next_stage_ids = [
        t.target_id for t in stage.workflow_transitions if t.type == transition_type
    ]

    return [
        dict(stage=stage_id, data={"item": item})
        for item in variable_value
        for stage_id in next_stage_ids
    ]


def action_strategy(stage: ActionStage, stage_run: StageRun, transition_type: str):
    if transition_type == "forms:abandoned" and not stage.is_initial:
        return [dict(stage=stage_run.stage, data=stage_run.data)]

    next_stage_ids = [
        t.target_id for t in stage.workflow_transitions if t.type == transition_type
    ]

    return [dict(stage=stage_id, data=stage_run.data) for stage_id in next_stage_ids]


def factory():
    return WorkflowEngine(get_stage_run_repository())


workflow_engine = factory()
